use once_cell::sync::OnceCell;
use paris::info;
use serde::{Deserialize, Serialize};
use std::{sync::Arc, time::Duration};
use tauri::async_runtime::RwLock;
use tokio::{
    task,
    time::{sleep, Instant},
};
use tracing::warn;

use crate::{
    picker::{led_color::LedColor, manager::Picker},
    rpc,
};

#[derive(Debug, Serialize, Deserialize)]
pub enum AmbientLightMode {
    None,
    Follow,
    Flowing,
}

pub struct CoreManager {
    ambient_light_mode: Arc<RwLock<AmbientLightMode>>,
}

impl CoreManager {
    pub fn global() -> &'static CoreManager {
        static CORE_MANAGER: OnceCell<CoreManager> = OnceCell::new();

        let core = CORE_MANAGER.get_or_init(|| CoreManager {
            ambient_light_mode: Arc::new(RwLock::new(AmbientLightMode::None)),
        });

        tokio::spawn(async {
            loop {
                core.play_flowing_light().await;
                match core.play_follow().await {
                    Ok(_) => {}
                    Err(error) => {
                        warn!("Can not following displays. {}", error);
                        sleep(Duration::from_millis(1000)).await;
                    }
                };
                sleep(Duration::from_millis(10)).await;
            }
        });
        core
    }

    pub async fn set_ambient_light(&self, target_mode: AmbientLightMode) {
        let mut mode = self.ambient_light_mode.write().await;
        *mode = target_mode;
    }

    pub async fn play_flowing_light(&self) {
        let mut hue = 0f64;
        let step_length = 2.0;
        loop {
            let lock = self.ambient_light_mode.read().await;
            if let AmbientLightMode::Flowing = *lock {
                let mut colors = Vec::<LedColor>::new();
                for i in 0..60 {
                    let color =
                        LedColor::from_hsv((hue + i as f64 * step_length) % 360.0, 1.0, 0.5);
                    colors.push(color);
                }
                hue = (hue + 1.0) % 360.0;
                match rpc::manager::Manager::global()
                    .publish_led_colors(&colors)
                    .await
                {
                    Ok(_) => {}
                    Err(error) => {
                        warn!("publish led colors failed. {}", error);
                    }
                }
            } else {
                break;
            }
            sleep(Duration::from_millis(50)).await;
        }
    }

    pub async fn play_follow(&self) -> anyhow::Result<()> {
        {
            let lock = self.ambient_light_mode.read().await;
            if let AmbientLightMode::Follow = *lock {
                Picker::global().refresh_displays().await?;
            } else {
                return Ok(());
            }
        };

        loop {
            let start = Instant::now();
            let next_tick = start + Duration::from_millis(16);
            info!("Following");
            let lock = self.ambient_light_mode.read().await;
            if let AmbientLightMode::Follow = *lock {
                task::spawn(async {
                    let start = Instant::now();
                    match Self::follow_once().await {
                        Ok(_) => {}
                        Err(error) => {
                            warn!("take screenshots failed. {}", error);
                        }
                    };
                    println!(
                        "Time elapsed in expensive_function() is: {:?}",
                        start.elapsed()
                    );
                });
            } else {
                break;
            }
            tokio::time::sleep_until(next_tick).await;
        }

        Ok(())
    }

    async fn follow_once() -> anyhow::Result<()> {
        Picker::global().take_screenshots_for_all().await?;
        let colors = Picker::global().get_led_strip_colors().await?;

        let colors = colors.into_iter().rev().collect();
        rpc::manager::Manager::global()
            .publish_led_colors(&colors)
            .await?;
        Ok(())
    }
}
